"""Get definitions for all words in GloVe vocab from dictionary.com"""

import json
import logging
import time
import typing
import requests
from bs4 import BeautifulSoup
from string import punctuation
from tqdm import tqdm
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

def extract_definitions(word):
    """Access dictionary.com and return a list
    of found definitions for the given word."""

    # format word for url
    word = word.replace("'", "-")

    # the website takes care of capitalisation issues issues on its own
    # e.g., dictionary.com/browse/germany and dictionary.com/browse/Germany both direct
    # to the same entry for Germany 
    url = f"https://www.dictionary.com/browse/{word}"

    session = requests.Session()
    retry = Retry(connect=3, backoff_factor=10)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    response = session.get(url)

    if response.status_code == 200:
        soup = BeautifulSoup(response.content, "html.parser")
        definition_divs = soup.find_all("div", {"data-type": "word-definition-content"})

        definitions = []
        for div in definition_divs:
            p_tags = div.find_all("p")
            for p_tag in p_tags:
                definitions.append(p_tag.get_text())
  
        return definitions
    else:
        logging.debug(f"Failed to fetch definitions for '{word}'. Status code: {response.status_code}")
        return []

def main(start: int=0, end: typing.Optional[int]=None):
    """Execute the crawling function for all words in the vocab,
    or, alternatively, specify a range in which vocab words should be considered.
    Save the definitions in a json file."""

    logging.info("Reading vocab ...")
    with open("/workspace/students/reichelt/BA/data/dd-glove/english_vocab.txt",
              "r", encoding="utf-8") as f:
        lines = f.readlines()
         # vocab.txt contains entries like "0 the" -> extract only the token without index
        word_list = [line.split()[0] for line in lines]
    logging.info("Reading done.")

    all_definitions = {}  # create empty dictionary to store definitions
    empty_returns = 0  # count tokens without defintion
    logging.info("Crawling definitons ...")
    for i, word in tqdm(enumerate(word_list[start:end])):  # TODO ideally log this instead
        # wait between requests to not overload their servers
        if i % 100 == 0:
            time.sleep(5)
        elif i % 20 == 0:
            time.sleep(1)
        definitions = extract_definitions(word)
        if not definitions:
            empty_returns += 1
        all_definitions[word] = definitions  # store the definitions in the dictionary
        logging.debug(f"Extracted {len(definitions)} definitions for {word}")
    logging.info(f"Crawling done. {empty_returns} out of {len(word_list[start:end])} tokens had no definiton.")

    # store the definitions in a json file
    with open("/workspace/students/reichelt/BA/data/dd-glove/english_definitions.json", "w") as f:
        logging.info("Saving definitions ...")
        json.dump(all_definitions, f, indent=4)
        logging.info("Saving done.")

if __name__ == "__main__":
    logging.basicConfig(filename='crawl_english_definitions.log', level=logging.INFO,
                        format='%(asctime)s - %(levelname)s: %(message)s', filemode="w")
    main()  # e.g. end=100 means use only first 100 vocab words
