"""Main content changed from original to account for German definitions
from GermaNet or DDUW (DUDEN Deutsches Universalwörterbuch), depending on specification

Notes:
- is called by the demo.sh script or run_definition.sh
- takes vocab list and desired definition source (germanet, dduw or dictionary.com) as input

If desired change:
ll. 67-70 if custom definitions desired
ll. 119, 122, 133, 134 for save file location
"""

import argparse
import json
import string
import nltk
import numpy as np
import pandas as pd
from tqdm import tqdm
from nltk.tokenize import WordPunctTokenizer

nltk.download("punkt")
nltk.download("perluniprops")

parser = argparse.ArgumentParser()
parser.add_argument('--vocab_file', default='vocab.txt', type=str)
parser.add_argument('--def_source', default='dduw', type=str)
parser.add_argument('--max_vocab_size', default=400000, type=int)
parser.add_argument('--stopwords_file', default='stopwords.txt', type=str)
parser.add_argument('--save_name', default='definitions', type=str)
args = parser.parse_args()

if args.def_source == 'germanet':
    from germanetpy.germanet import Germanet

extended_punctuation = string.punctuation + "„“‚‘»«–—…||´´``''\"\""

def use_germanet(def_counter, ambig_counter):
    germanet = Germanet("/workspace/students/reichelt/BA/data/Germanet")
    for idx, word in tqdm(enumerate(train_words)):
        synsets = germanet.get_synsets_by_orthform(word)
        if not synsets:
            synsets = germanet.get_synsets_by_orthform(word.capitalize())
        for synset in synsets:
            for lexunit in synset.lexunits:
                orthforms = lexunit.get_all_orthforms()
                orthforms = [o.lower() for o in orthforms]
                if word in orthforms:
                    definitions = lexunit.wiktionary_paraphrases
                    if definitions:
                        def_counter += 1
                    if len(definitions) > 1:
                        ambig_counter += 1
                    for definition in definitions:
                        def_text = definition.wiktionary_sense
                        for def_w in tokenizer.tokenize(def_text):
                            def_w = def_w.lower()
                            if def_w != word and def_w in train_words and not def_w in stopwords and not def_w in extended_punctuation:
                                def_history[idx].append(word_to_id[def_w])
    return def_counter, ambig_counter

def use_dduw(def_counter, ambig_counter):
    dduw = pd.read_csv("/workspace/students/reichelt/BA/data/dduw.csv", sep=";")
    for idx, word in tqdm(enumerate(train_words)):
        relevant_entries = dduw[(dduw["Lemma"] == word) | (dduw["Lemma"] == word.capitalize())]  # train data and vocab treat tokens case-insensitive, therefore we want definitions for both cases
        if not relevant_entries.empty:
            def_counter += 1
            definitions = relevant_entries["Definition/Erklärung"].tolist()
            if len(definitions) > 1:
                ambig_counter += 1
            definitions = [str(s) for s in definitions]  # there are some definitions which apparently look like floats
            concatenated_defs = " ".join(definitions)
            #if word == "mann":
            #    concatenated_defs = "Mann Person männlichen Geschlechts Junge männlich Geschlecht er sein seine seinem ihm Ehemann Hengst Bruder Mädel weibliche weibliches weiblichem Frauen"
            #elif word == "frau":
            #    concatenated_defs = "Frau Person weiblichen Geschlechts Mädchen weiblich Geschlecht sie ihr ihre ihrem ihr Ehefrau Stute Schwester Knabe männliche männliches männlichem Männer"
            for def_w in tokenizer.tokenize(concatenated_defs):
                def_w = def_w.lower()
                if def_w != word and def_w in train_words and not def_w in stopwords and not def_w in extended_punctuation:
                    def_history[idx].append(word_to_id[def_w])
    return def_counter, ambig_counter


def use_dictionary_com(def_counter, ambig_counter):
    with open("/workspace/students/reichelt/BA/data/dd-glove/english_definitions.json",
              "r", encoding="utf-8") as json_file:
        data = json.load(json_file)
    
    # Since dictionaries are inherently unordered, iterate over the vocab instead
    # of just going through the json file even though it contains the same 400,000 tokens 
    for i, current_token in tqdm(enumerate(train_words)):
        try:
            defs = data[current_token]
            if defs:
                def_counter += 1  # word has definitions stored, keep track
            if len(defs) > 1:
                ambig_counter += 1  # word has multiple definitions, keep track
            defs = [str(s) for s in defs]  # to prevent type issues
            concatenated_defs = " ".join(defs)
            tokenized_defs = tokenizer.tokenize(concatenated_defs)
            for tok in tokenized_defs:
                tok = tok.lower()
                if tok != current_token and tok in train_words and not tok in stopwords and not tok in extended_punctuation:
                    def_history[i].append(word_to_id[tok])
        except KeyError:
            print(f"NOT ABLE TO FIND ENTRY FOR '{current_token}'!")
            continue
    return def_counter, ambig_counter

train_words = []
word_to_id = {}

print('Reading from vocab file...')
with open(args.vocab_file, 'r', encoding='utf-8') as f:
    for idx, line in enumerate(f.readlines()):
        train_words.append((line.strip().split()[0]))
        if len(train_words) == args.max_vocab_size:
            break

word_to_id = {word: i for i, word in enumerate(train_words)}

print('check length of word lists: {}'.format(len(train_words)))  # = vocab size

stopwords = set()
with open(f'{args.stopwords_file}', encoding='utf-8') as f:
    for line in f.readlines():
        stopwords.add(line.strip())

print('Done reading stopwords file.')

def_history = {}
for idx, word in enumerate(train_words):
    def_history[idx] = []

tokenizer = WordPunctTokenizer()

definition_counter = 0
ambiguous_counter = 0

if args.def_source == 'germanet':
    definition_counter, ambiguous_counter = use_germanet(definition_counter, ambiguous_counter)
elif args.def_source == 'dduw':
    definition_counter, ambiguous_counter = use_dduw(definition_counter, ambiguous_counter)
elif args.def_source == 'dictionary_com':
    definition_counter, ambiguous_counter = use_dictionary_com(definition_counter, ambiguous_counter)
else:
    raise ValueError('Invalid definition source specified.')

print(f'check length of def_history: {len(def_history)}')
to_write = []
file_obj = open(f'/workspace/students/reichelt/BA/data/dd-glove/{args.save_name}.dat', 'wb')
to_write.append(len(def_history))

print(f'Writing to {args.save_name}.dat ...')

for w_id, def_ids in def_history.items():
    to_write.append(w_id)
    to_write.append(len(def_ids))
    to_write.extend(def_ids)
np.array(to_write, dtype=np.int32).tofile(file_obj)
file_obj.close()

print('done')

with open(f'/workspace/students/reichelt/BA/data/dd-glove/{args.save_name}.txt', 'w') as f:
    print(f'Writing to {args.save_name}.txt ...')
    for word_id, definitions in def_history.items():
        content = '{}\t{}\n'.format(word_id, definitions)
        f.write(content)

print('done')

print(f'found at least 1 definition for {definition_counter} token(s) out of {len(train_words)}')
print(f'out of those, {ambiguous_counter} token(s) had multiple definitions')

