"""Helper script to get seed words, their
definitions, and their definition's embedding's bias
from a list of seed word indexes.

'Girl' and 'Boy' are used as placeholder terms for the
two groups, can also mean e.g. German and Foreign."""

import ast
import json
import pandas as pd
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity as cosine

# hard-coded, you can get seed word indexes from GloVe training output
girl_indexes = [
    41, 2933,   624,  7595,  2626,  6107,  5306,   807, 10470,   125,  4877,
        7744,  8440,   425,  2068,   971,  3026,  1674,   618,  3180,  2236,
         428,  1780,  2484,    42,  3237,  4977,  1417,  1067,  3819,  1452
    ]
boy_indexes = [
    16, 98915,  65269,  42557, 138164,  97023,  18330, 134737,  27996,  57301,
        17621,  14234,  47588, 108234, 211932,  77434,  37725,  47421, 346084,
       303133,  47821,  22468,  44522, 330544,  79490,  49760, 109485, 214051,
        47483,  53517,  89180
    ]

# dicts will contain pairs of tokens and their dictionary definitions
girl_words = dict()
boy_words = dict()

# look up tokens corresponding to indexes in vocab text file
with open("/workspace/students/reichelt/BA/data/dd-glove/english_vocab.txt", encoding="utf-8") as vocab:
    for idx, line in enumerate(vocab.readlines()):
        token = line.split()[0]
        if idx in girl_indexes:
            girl_words[idx] = dict() # {10: {"token": mann, "definition": "...", "bias": 0.0}}
            girl_words[idx]["token"] = token
        if idx in boy_indexes:
            boy_words[idx] = dict()
            boy_words[idx]["token"] = token

with open("/workspace/students/reichelt/BA/data/dd-glove/english_definitions.json",
          "r", encoding="utf-8") as f:
    data = json.load(f)
for idx, info in girl_words.items():
    word = info["token"]
    girl_words[idx]["definition"] = data[word]
for idx, info in boy_words.items():
    word = info["token"]
    boy_words[idx]["definition"] = data[word]

# calculate bias of seed words and sort accordingly

# load embeddings
embeddings = pd.read_csv("/workspace/students/reichelt/BA/data/dd-glove/english_vectors_gender_1iter.txt",
                         skiprows=1, header=None, sep=" ")
embeddings.rename(columns={0: "token"}, inplace=True)  # name first column "token"
embeddings["vector"] = embeddings.iloc[:, 1:].values.tolist()  # convert other columns into one, containing 300-dim vec as a list

# load definition indexes
definition_indexes = pd.read_csv("/workspace/students/reichelt/BA/data/dd-glove/english_definitions.txt", sep="\t", usecols=[1], header=None, names=["def_words"])
definition_indexes = definition_indexes["def_words"].tolist()
definition_indexes = [ast.literal_eval(item) for item in definition_indexes]

def get_definition_embedding(word_index: int) -> np.array:
    """Calculate definition embedding by averaging embeddings of
    words occurring in given definition. Definition is given by
    a word index, i.e. 88 which is token 'Mann'."""
    emb_sum = np.zeros(300)
    if definition_indexes[word_index]:
        for i in definition_indexes[word_index]:
            vec = np.array(embeddings["vector"].iloc[i])  # get embedding of token with index idx
            emb_sum += vec
        definitional_embedding = (1/len(definition_indexes[word_index])) * emb_sum
        return definitional_embedding
    return emb_sum

def calculate_definitional_bias(v_1: np.array, v_2: np.array, w: int) -> float:
    """Calculate bias of a word using is definitional embedding.
    Calculate the projection of the word's definitional embedding
    along the difference between the seed words' definitional embeddings."""
    boy_def = v_1.reshape(1, -1)  # reshape to fit cosine function
    girl_def = v_2.reshape(1, -1)
    word_def = get_definition_embedding(w).reshape(1, -1)
    bias = cosine(word_def, boy_def) - cosine(word_def, girl_def)
    return bias

# look up def embedding for initial seed words
mann_vec = get_definition_embedding(16)
frau_vec = get_definition_embedding(41)

# calculate girl biases and sort in ascending order (= strongest bias first, here)
for girl_index in girl_indexes:
    b = calculate_definitional_bias(mann_vec, frau_vec, girl_index)
    girl_words[girl_index]["bias"] = b[0].item()

girl_words_sorted = dict(sorted(girl_words.items(), key=lambda item: item[1]["bias"]))

# calculate boy biases and sort in descending order (= strongest bias first, here)
for boy_index in boy_indexes:
    b = calculate_definitional_bias(mann_vec, frau_vec, boy_index)
    boy_words[boy_index]["bias"] = b[0].item()

boy_words_sorted = dict(sorted(boy_words.items(), key=lambda item: item[1]["bias"], reverse=True))

# write results to files
with open("/home/students/reichelt/ba/bias-mitigation-ba/DD-GloVe/seed_word_tests/she_seedwords_alternate_calc.txt",
          "w", encoding="utf-8") as f:
    for idx, info in girl_words_sorted.items():
        f.write(info["token"].upper() + "\n")
        for d in info["definition"]:
            f.write(f"Def.: {d}\n")
        f.write(f"Definitional bias: {info['bias']:.4f}\n\n")

with open("/home/students/reichelt/ba/bias-mitigation-ba/DD-GloVe/seed_word_tests/he_seedwords_alternate_calc.txt",
          "w", encoding="utf-8") as f:
    for idx, info in boy_words_sorted.items():
        f.write(info["token"].upper() + "\n")
        for d in info["definition"]:
            f.write(f"Def.: {d}\n")
        f.write(f"Definitional bias: {info['bias']:.4f}\n\n")
