"""Determine seed words as described by algorithm in the paper"""

import ast
import pandas as pd
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity as cosine
from tqdm import tqdm

# load embeddings
embeddings = pd.read_csv("/workspace/students/reichelt/BA/data/dd-glove/english_vectors_gender.txt",
                         skiprows=1, header=None, sep=" ")
embeddings.rename(columns={0: "token"}, inplace=True)  # name first column "token"
embeddings["vector"] = embeddings.iloc[:, 1:].values.tolist()  # convert other columns into one, containing 300-dim vec as a list

# load definition indexes
definition_indexes = pd.read_csv("/workspace/students/reichelt/BA/data/dd-glove/english_definitions.txt", sep="\t", usecols=[1], header=None, names=["def_words"])
definition_indexes = definition_indexes["def_words"].tolist()
definition_indexes = [ast.literal_eval(item) for item in definition_indexes]

def get_definition_embedding(word_index: int) -> np.array:
    """Calculate definition embedding by averaging embeddings of
    words occurring in given definition. Definition is given by
    a word index, i.e. 88 which is token 'mann'."""
    emb_sum = np.zeros(300)
    try:
        for i in definition_indexes[word_index]:  # look at all definitional words
            vec = np.array(embeddings["vector"].iloc[i])  # get embedding of a definitional word
            emb_sum += vec  # add to sum of all definitional embeddings
    except TypeError as exc:
        print(word_index)
        print(definition_indexes[word_index])
        raise TypeError from exc
    definitional_embedding = (1/len(definition_indexes[word_index])) * emb_sum  # build mean
    return definitional_embedding

def calculate_definitional_bias(v_1: np.array, v_2: np.array, w: np.array) -> float:
    """Calculate bias of a word using is definitional embedding.
    Calculate the projection of the word's definitional embedding
    along the difference between the seed words' definitional embeddings."""
    boy_def = v_1.reshape(1, -1)  # reshape to fit sklearn's cosine function
    girl_def = v_2.reshape(1, -1)
    word_def = w.reshape(1, -1)
    bias = cosine(word_def, boy_def) - cosine(word_def, girl_def)
    return bias[0].item()  # because of how sklearn's cosine works

def alternate_bias(v_1: np.array, v_2: np.array, w: np.array) -> float:
    """Alternatively try out using actual projection as described
    textually in the paper instead of cosine - cosine.
    DON'T convert w to a unit vector through division by ||w||.
    Result: bias = w * (v1-v2) / ||(v1-v2)||
    """
    #unit_w = w / np.linalg.norm(w) 
    diff = v_1 - v_2
    dot_product = np.dot(w, diff)
    norm = np.linalg.norm(diff)
    return dot_product / norm

# look up def embedding for initial seed words
mann_vec = get_definition_embedding(16)
frau_vec = get_definition_embedding(41)

# for each word in vocab, calculate bias and save results in vocab dataframe
def def_bias_application(vector_column_value):
    emb = np.array(vector_column_value)  # dataframe contains vectors as lists in column "vector"
    b = alternate_bias(mann_vec, frau_vec, emb)
    return b

# apply calculation to whole table with progress bar
tqdm.pandas()
embeddings["definitional_bias"] = embeddings["vector"].progress_apply(def_bias_application)

# get top 10 highest and lowest indices (as described in paper)
boy_indices = embeddings["definitional_bias"].nlargest(30).index
girl_indices = embeddings["definitional_bias"].nsmallest(30).index

print(f"boy indices: {boy_indices}")
print(f"girl indices {girl_indices}")
