"""
WordSim-353 score calculation for a given pair of embeddings.
Example usage: python evaluate.py --version gur350 --vector_location /path/to/vec
"""
import argparse
import pandas as pd
from gensim.models import KeyedVectors
from scipy.stats import spearmanr
from scipy.spatial.distance import cosine

def load_word_lists(version: str):
    """Loads word pair scores as annotated by humans and published.
    Different versions:
    - english = original scores spit into relatedness and similarity
    - english_complete = same as english, but includes the duplicate money-cash
      entry, which the authors of the rel/sim split removed
    - english_multilingual = excludes three hard to transalte pairs and has
      has new scores, annotated in same study as other translated data
    - german = translated word pairs, annotated in same study as english_multilingual
    - german = different version, annotated by Gurevych et al. 2005, 350 pairs"""
    
    if version == "english":
        file_path_rel = "/home/students/reichelt/ba/bias-mitigation-ba/WordSim353/data/ws353simrel_agirre/wordsim_relatedness_goldstandard.txt"
        file_path_sim = "/home/students/reichelt/ba/bias-mitigation-ba/WordSim353/data/ws353simrel_agirre/wordsim_similarity_goldstandard.txt"
        correct_length = 352  # excludes the duplicate entry money-cash
        header = None
    elif version == "english_complete":
        file_path_rel = "data/ws353simrel_agirre/wordsim_relatedness_goldstandard.txt"
        file_path_sim = "data/ws353simrel_agirre/wordsim_similarity_goldstandard_353.txt"
        correct_length = 353  # money-cash 9.08 added to similarity
        header = None
    elif version == "english_multilingual":
        file_path_rel = "data/multilingual_leviant_reichhart/wordsim353-english-rel.txt"
        file_path_sim = "data/multilingual_leviant_reichhart/wordsim353-english-sim.txt"
        correct_length = 350  # excludes some non-translatable entries
        header = 0  # meaning file has a header in first line
    elif version == "german":
        file_path_rel = "data/multilingual_leviant_reichhart/wordsim353-german-rel.txt"
        file_path_sim = "data/multilingual_leviant_reichhart/wordsim353-german-sim.txt"
        correct_length = 350
        header = 0
    elif version == "gur350":
        file_path_rel = "data/german_gurevych_zesch/gur_350.tsv"  # only combined available
        file_path_sim = "data/german_gurevych_zesch/gur_350.tsv"  # only combined available
        correct_length = 350
        header = 0
    elif version == "gur283":
        file_path_rel = "data/german_gurevych_zesch/gur_350_reduced.tsv"  # only combined available
        file_path_sim = "data/german_gurevych_zesch/gur_350_reduced.tsv"  # only combined available
        correct_length = 283
        header = 0
    else:
        raise ValueError("invalid language/word lists specified")
    
    rel_df = pd.read_csv(file_path_rel,
                            encoding="utf-8", header=header, sep="\t", usecols=[0,1,2],
                            names=["Word 1", "Word 2", "Human Relatedness"])
    sim_df = pd.read_csv(file_path_sim,
                         encoding="utf-8", sep="\t", header=header, usecols=[0,1,2],
                         names=["Word 1", "Word 2", "Human Similarity"])

    # Split into similarity and relatedness doesn't change scores, so
    # the tables can be combined back into one for easier computation later on
    merged_df = sim_df.merge(rel_df, on=["Word 1", "Word 2"], how="outer")
    merged_df["Combined Human Score"] = merged_df["Human Similarity"].fillna(
        merged_df["Human Relatedness"])

    # Check data
    # Combined Score should be filled everywhere
    assert merged_df['Combined Human Score'].notna().all()

    # Sim and Rel should have same values where both are filled
    assert merged_df[(merged_df['Human Similarity'].notna()) 
                     & (merged_df['Human Relatedness'].notna())]['Human Similarity'].equals(
                     merged_df[(merged_df['Human Similarity'].notna()) 
                     & (merged_df['Human Relatedness'].notna())]['Human Relatedness'])
    
    # There should be 353 rows
    assert len(merged_df) == correct_length

    return merged_df


def calculate_wordsim_score(word_pair_scores: pd.DataFrame, file_path: str) -> dict:
    """Loads word vectors, calculates cosine similiarity scores for word pairs
    and compares with human scores using Spearman Rank Correlation"""
    
    # Load word vectors
    if "fasttext" in file_path:
        #model = FastText.load_fasttext_format(file_path)
        #model.init_sims(replace=True)
        model = KeyedVectors.load_word2vec_format(file_path, binary=False)
        model.init_sims(replace=True)
    elif file_path[-4:] == ".bin":
        model = KeyedVectors.load_word2vec_format(file_path, binary=True)
        model.init_sims(replace=True)
    else:
        model = KeyedVectors.load_word2vec_format(file_path, binary=False)
        model.init_sims(replace=True)

    # Calculate embedding similarites and add to df as new column
    def calculate_embedding_similarity(row):
        word_1 = row["Word 1"].lower()
        word_2 = row["Word 2"].lower()

        if word_1 in model:
            embd_1 = model[word_1]
        elif repr(word_1.encode("utf-8")) in model:
            embd_1 = model[repr(word_1.encode("utf-8"))]  # deepset models have weird format
        else:
            print(f"{word_1} not in vocab. returning similarity 0.")
            return 0.0
        
        if word_2 in model:
            embd_2 = model[word_2]
        elif repr(word_2.encode("utf-8")) in model:
            embd_2 = model[repr(word_2.encode("utf-8"))]
        else:
            print(f"{word_2} not in vocab. returning similarity 0.")
            return 0.0

        return 1 - cosine(embd_1, embd_2)  # scipy implements cosine distance not similarity


    word_pair_scores["Embedding Score"] = word_pair_scores.apply(
        calculate_embedding_similarity, axis=1)

    # Calculate correlations
    combined_rho, combined_p = spearmanr(
        word_pair_scores["Combined Human Score"], word_pair_scores["Embedding Score"],
        nan_policy="raise"
    )

    sim_rows = word_pair_scores.dropna(subset=['Human Similarity', 'Embedding Score'])
    rel_rows = word_pair_scores.dropna(subset=['Human Relatedness', 'Embedding Score'])

    sim_rho, sim_p = spearmanr(
        sim_rows["Human Similarity"], sim_rows["Embedding Score"], nan_policy="raise"
    )
    rel_rho, rel_p = spearmanr(
        rel_rows["Human Relatedness"], rel_rows["Embedding Score"], nan_policy="raise"
    )

    # For easier access later on
    results = {
        "sim_rho": sim_rho, "sim_p": sim_p, "rel_rho": rel_rho,
        "rel_p": rel_p, "combined_rho": combined_rho, "combined_p": combined_p
        }

    return results


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Calculate wordsim353 score for given language and vectors")
    parser.add_argument("--version", default="english",
                        help="'english', 'english_complete', 'english_multilingual', 'german', 'gur350'")
    parser.add_argument("--vector_location", help="specify a file path to embeddings")
    args = parser.parse_args()

    print("Loading data ...")
    data = load_word_lists(args.version)
    print("Loading successful. Calculating score ...")
    scores = calculate_wordsim_score(data, args.vector_location)
    output = (
        "-----------------------------\n"
        f"WordSim-353 scores (Spearman's rank correlation) for '{args.version}' word lists and embeddings from {args.vector_location}:\n"
        f"Similarity: {scores['sim_rho']:.4f} with p-value {scores['sim_p']:.5f}\n"
        f"Relatedness: {scores['rel_rho']:.4f} with p-value {scores['rel_p']:.5f}\n"
        f"Combined: {scores['combined_rho']:.4f} with p-value {scores['combined_p']:.5f}\n"
        "-----------------------------"
    )
    print(output)
