"""
Hard-debias embedding

Man is to Computer Programmer as Woman is to Homemaker? Debiasing Word Embeddings
Tolga Bolukbasi, Kai-Wei Chang, James Zou, Venkatesh Saligrama, and Adam Kalai
2016
"""
from __future__ import print_function, division
import argparse
import json
import numpy as np
# we = __import__("debiaswe-master.debiaswe.we")
import we  # linter isn't happy but it works, unlike attempt above
import gensim
import fasttext

def debias(E, gender_specific_words, definitional, equalize):
    gender_direction = we.doPCA(definitional, E).components_[0]
    specific_set = set(gender_specific_words)
    for i, w in enumerate(E.words):
        if w not in specific_set:
            E.vecs[i] = we.drop(E.vecs[i], gender_direction)
    E.normalize()
    candidates = {x for e1, e2 in equalize for x in [(e1.lower(), e2.lower()),
                                                     (e1.title(), e2.title()),
                                                     (e1.upper(), e2.upper())]}
    print(candidates)
    for (a, b) in candidates:
        if (a in E.index and b in E.index):
            y = we.drop((E.v(a) + E.v(b)) / 2, gender_direction)
            z = np.sqrt(1 - np.linalg.norm(y)**2)
            if (E.v(a) - E.v(b)).dot(gender_direction) < 0:
                z = -z
            E.vecs[E.index[a]] = z * gender_direction + y
            E.vecs[E.index[b]] = -z * gender_direction + y
    E.normalize()

def remove_oov(word_list: list, vocab: list) -> list:
    """Takes a definitional, equalisation or bias-specific word list
    and removes those words which are out of the model's vocabulary.
    Relevant especially for the GloVe model. Reports on removal."""
    
    cleaned_list = []
    
    for element in word_list:
        if type(element) == list:
            if element[0] in vocab and element[1] in vocab:
                cleaned_list.append(element)
            else:
                print(f"Removed element {element}")
        else:
            if element in vocab:
                cleaned_list.append(element)
            else:
                print(f"Removed element {element}")
    
    return cleaned_list


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("embedding_filename", help="The name of the embedding")
    parser.add_argument("definitional_filename", help="JSON of definitional pairs")
    parser.add_argument("gendered_words_filename", help="File containing words not to neutralize (one per line)")
    parser.add_argument("equalize_filename", help="???.bin")
    parser.add_argument("debiased_filename", help="???.bin")

    args = parser.parse_args()
    print(args)

    with open(args.definitional_filename, "r") as f:
        defs = json.load(f)

    with open(args.equalize_filename, "r") as f:
        equalize_pairs = json.load(f)

    with open(args.gendered_words_filename, "r") as f:
        gender_specific_words = json.load(f)
    
    if args.embedding_filename.endswith("wiki.de.bin"):
        ft = fasttext.load_model(args.embedding_filename)
        words = ft.words
    elif args.embedding_filename.endswith("vectors_no_debiasing.txt"):
        model = gensim.models.KeyedVectors.load_word2vec_format(args.embedding_filename, binary=False)
        words = model.index_to_key

    defs = remove_oov(defs, words)
    equalize_pairs = remove_oov(equalize_pairs, words)
    gender_specific_words = remove_oov(gender_specific_words, words)

    E = we.WordEmbedding(args.embedding_filename)

    print("Debiasing...")
    debias(E, gender_specific_words, defs, equalize_pairs)

    print("Saving to file...")
    if args.embedding_filename[-4:] == args.debiased_filename[-4:] == ".bin":
        E.save_w2v(args.debiased_filename)
    else:
        E.save(args.debiased_filename)

    print("\n\nDone!\n")
